// dobble.c (RECURSIVE)
// Alain Brobecker - 2010/08/13

// S=NbSymbolsPerCard
// O=OccurencePerSymbol
// C=NbCards
// N=NbSymbols

#include <stdio.h>  /* file functions, etc... */
#include <stdlib.h> /* exit(), malloc() */

#define TRUE 1
#define FALSE 0

#define MaxNbCards 20*20
#define MaxNbSymbols 400


int NbSymbolsPerCard,OccurencePerSymbol; // read them from the command line
int NbCards,NbSymbols,NbIdenticalSymbols;
int Card[MaxNbCards][MaxNbSymbols];
int SymbolOccurence[MaxNbSymbols];
int CombinationPosition[MaxNbCards][MaxNbSymbols];


/**** ReadNumber ****/
int ReadNumber(s) char *s; {
  int n,t,sign;
  unsigned char c;
  n=0;
  sign=1;
  t=0;
  c=s[t];
  if(c=='-') { sign=-1; t++; c=s[t]; }
  while(c!=0) {
    if((c<'0') || (c>'9')) {
      printf("Error: %s should be a number",s);
      exit(1);
    }
    n=n*10+c-'0';
    t++;
    c=s[t];
  }
  return(n*sign);
}

/**** All symbols appear the correct number of times? ****/
int AllSymbolsOccurCorrectly() {
  int c=0;
  while((c<NbSymbols) && (SymbolOccurence[c]==OccurencePerSymbol)) { c++; }
  if(c==NbSymbols) { return(TRUE); }
  return(FALSE);
}

/**** Count the number of identical symbols between two cards ****/
int CountIdenticalSymbols(int Card1,int Card2) {
  int d,nb;
  nb=0;
  for(d=0;d<NbSymbols;d++) {
    if((Card[Card1][d]!=FALSE) && (Card[Card1][d]==Card[Card2][d])) { nb++; }
  }
  return(nb);
}

/**** Print the cards ****/
void PrintCards() {
  int c,d;
  for(c=0;c<NbCards;c++) {
    for(d=0;d<NbSymbols;d++) {
      if(Card[c][d]==FALSE) {
        printf(".");
      } else {
        if(d<26) { printf("%c",d+'a'); } else { printf("%c",d-26+'A'); }
      }
    }
    printf("\n");
  }
  printf("\n");
}

//**** RECURSIVE COMBINATIONS ****
//S=3 NbSymbolsPerCard
//O=2 OccurencePerSymbol
//C=4 NbCards
//N=6 NbSymbols
//Combinations of S-1=2 elements amongst N-S=6-3, starting at position S
// abc...
// a..de.
// .b.###
// ..c###

//!!! void CardCombinations(int CardNumber,int Variable,int start,int MaxValue, int NbVariables) {

    
//Recursive setup of all nCpcombinations for a given card
//Call it with "Combinations(CardNumber,0,0,n-p);"
//Here n=NbSymbols-NbSymbolsPerCard
// and p=NbSymbolsPerCard-1
void CardCombinations(int CardNumber,int Variable,int start,int end) {
  int c,d,e,boolean;
  for(c=start;c<=end;c++) {
    CombinationPosition[CardNumber][Variable]=NbSymbolsPerCard+c;
    if(Variable==NbSymbolsPerCard-1-1) {
      for(d=NbSymbolsPerCard;d<NbSymbols;d++) { Card[CardNumber][d]=FALSE; }
      for(d=0;d<NbSymbolsPerCard-1;d++) { Card[CardNumber][CombinationPosition[CardNumber][d]]=TRUE; }

      //Check if the Card follows the constraints
      //First constraint: only one identical symbol between two cards
      boolean=TRUE;
      d=0;
      while((d<CardNumber) && (boolean==TRUE)) {
        if(CountIdenticalSymbols(d,CardNumber)!=1) { boolean=FALSE; }
        d++;
      }
      //Second constraint: no symbol appears more than OccurencePerSymbol times
      //We don't need to check the first SymbolPerCards symbols
      if(boolean==TRUE) {
        
//!!! PLUTOT FAIRE PAR COLONNE
        
        //count occurences
        for(e=NbSymbolsPerCard;e<NbSymbols;e++) { SymbolOccurence[e]=0; }
        for(d=1;d<=CardNumber;d++) {
          for(e=NbSymbolsPerCard;e<NbSymbols;e++) {
            if(Card[d][e]==TRUE) { SymbolOccurence[e]++; }
          }
        }
        e=NbSymbolsPerCard;
        while((e<NbSymbols) && (boolean==TRUE)) {
          if(SymbolOccurence[e]>OccurencePerSymbol) { boolean=FALSE; }
          e++;
        }
      }
      
      //This card is ok
      if(boolean==TRUE) {
        if(CardNumber==NbCards-1) {
          //It's the last card, then print
          //?We must still check that every symbol appears OccurencePerSymbol times?    
          PrintCards();
          exit(0);
        } else {
          //Not the last card, so process the remaining ones
          CardCombinations(CardNumber+1,0,0,NbSymbols-NbSymbolsPerCard-(NbSymbolsPerCard-1));

        }
       }
       
    } else {
      CardCombinations(CardNumber,Variable+1,c+1,end+1);
    }
  }
}


/**** MAIN *****************************************************************/
int main(argc,argv) int argc; char *argv[]; {
  setbuf(stdout, NULL); /* no buffering for output! */
  int c,d,e;
  
  //read parameters from command line
  if(argc!=3) { printf("Usage: dobble NbSymbolsPerCard OccurencePerSymbol\n"); exit(1); }
  NbSymbolsPerCard=ReadNumber(argv[1]);
  OccurencePerSymbol=ReadNumber(argv[2]);
  NbCards=1+NbSymbolsPerCard*(OccurencePerSymbol-1);
  NbSymbols=NbCards*NbSymbolsPerCard/OccurencePerSymbol;
  printf("%d symbols per card\n",NbSymbolsPerCard);
  printf("%d occurences per symbol\n",OccurencePerSymbol);
  printf("%d cards\n",NbCards);
  printf("%d symbols\n",NbSymbols);

  //Empty tables
  for(d=0;d<MaxNbSymbols;d++) {
    for(c=0;c<MaxNbCards;c++) {
      Card[c][d]=FALSE;
    }
    SymbolOccurence[d]=0;
  }
  
  //Fill the O first lines, eg, S=4, O=3:
  // abcd........
  // a...efg.....
  // a......hij..
  for(c=0;c<OccurencePerSymbol;c++) {
    Card[c][0]=TRUE;
    e=1+c*(NbSymbolsPerCard-1);
    for(d=0;d<NbSymbolsPerCard-1;d++) {
      Card[c][e+d]=TRUE;
    }
  }
  
  //Fill the S first symbols of every line, eg, S=4, O=3:
  // ............
  // a...........
  // a...........
  // .b..........
  // .b..........
  // ..c.........
  // ..c.........
  // ...d........
  // ...d........
  for(c=0;c<NbSymbolsPerCard;c++) {
    e=1+c*(OccurencePerSymbol-1);
    for(d=0;d<OccurencePerSymbol-1;d++) {
      Card[e+d][c]=TRUE;
    }
  }

  //actualise the current occurence of symbols
  for(c=0;c<NbSymbolsPerCard;c++) { SymbolOccurence[c]=OccurencePerSymbol; }
  for(c=NbSymbolsPerCard;c<NbSymbols;c++) {SymbolOccurence[c]=1; }
  
  //There remains a block of (C-O) cards to complete,
  //with combinations of S-1 elements chosen amongst N-S,
  //and of course satisfying the conditions.  eg, S=4, O=3:
  // abcd........
  // a...efg.....
  // a......hij..
  // .b..########
  // .b..########
  // ..c.########
  // ..c.########
  // ...d########
  // ...d########
  
  
//Recursive setup of all nCpcombinations for a given card
//Call it with "Combinations(CardNumber,0,0,n-p);"
//Here n=NbSymbols-NbSymbolsPerCard
// and p=NbSymbolsPerCard-1
  CardCombinations(OccurencePerSymbol,0,0,NbSymbols-NbSymbolsPerCard-(NbSymbolsPerCard-1));
      
//  PrintCards();
  
  exit(0);
}
