// dobble.c
// Alain Brobecker - 2010/08/03

#include <stdio.h>  /* file functions, etc... */
#include <stdlib.h> /* exit(), malloc() */

#define TRUE 1
#define FALSE 0

#define MaxNbCards 10*10
#define MaxNbSymbols 100

int SymbolsPerCard,OccurencePerSymbol; // read them from the command line
int NbCards,NbSymbols,NbIdenticalSymbols;
int Card[MaxNbCards][MaxNbSymbols];
int SymbolOccurence[MaxNbSymbols];

/**** ReadNumber ****/
int ReadNumber(s) char *s; {
  int n,t,sign;
  unsigned char c;
  n=0;
  sign=1;
  t=0;
  c=s[t];
  if(c=='-') { sign=-1; t++; c=s[t]; }
  while(c!=0) {
    if((c<'0') || (c>'9')) {
      printf("Error: %s should be a number",s);
      exit(1);
    }
    n=n*10+c-'0';
    t++;
    c=s[t];
  }
  return(n*sign);
}

/**** All symbols appear the correct number of times? ****/
int AllSymbolsOccurCorrectly() {
  int c=0;
  while((c<NbSymbols) && (SymbolOccurence[c]==OccurencePerSymbol)) { c++; }
  if(c==NbSymbols) { return(TRUE); }
  return(FALSE);
}

/**** Count the number of identical symbols between two cards ****/
int CountIdenticalSymbols(int Card1,int Card2) {
  int d,nb;
  nb=0;
  for(d=0;d<NbSymbols;d++) {
    if((Card[Card1][d]!=FALSE) && (Card[Card1][d]==Card[Card2][d])) { nb++; }
  }
  return(nb);
}

/**** Print the cards ****/
void PrintCards() {
  int c,d;
  for(c=0;c<NbCards;c++) {
    for(d=0;d<NbSymbols;d++) {
      if(Card[c][d]==FALSE) {
        printf(".");
      } else {
        if(d<26) { printf("%c",d+'a'); } else { printf("%c",d-26+'A'); }
      }
    }
    printf("\n");
  }
  printf("\n");
}

/**** MAIN *****************************************************************/
int main(argc,argv) int argc; char *argv[]; {
  setbuf(stdout, NULL); /* no buffering for output! */
  int c,d,e,boolean;
  
  if(argc!=3) { printf("Usage: dobble SymbolsPerCard OccurencePerSymbol\n"); exit(1); }
  SymbolsPerCard=ReadNumber(argv[1]);
  OccurencePerSymbol=ReadNumber(argv[2]);
  
  //Empty tables
  for(d=0;d<MaxNbSymbols;d++) {
    for(c=0;c<MaxNbCards;c++) {
      Card[c][d]=FALSE;
    }
    SymbolOccurence[d]=0;
  }
  
  //Set first card to SymbolsPerCard first symbols
  for(d=0;d<SymbolsPerCard;d++) {
    Card[0][d]=TRUE;
    SymbolOccurence[d]=1;
  }
  NbCards=1;
  NbSymbols=SymbolsPerCard;
//!!!  PrintCards();
  
  while(    (NbCards<MaxNbCards)
         && (NbSymbols<MaxNbSymbols)
         && (AllSymbolsOccurCorrectly()==FALSE)) {
    //create a new card with one & only one symbol common to every other card
    for(c=0;c<NbCards;c++) {
      NbIdenticalSymbols=CountIdenticalSymbols(c,NbCards);
//!!!      printf("card %d and %d:%d identical symbols\n",c,NbCards,NbIdenticalSymbols);
      if(NbIdenticalSymbols>1) { printf("error: >1 identical symbols\n"); NbCards++; PrintCards(); exit(-1); } //!!!Useless???
      if(NbIdenticalSymbols==0) {
        //search first symbol not completly used in this card
        d=-1;
        boolean=FALSE;
        while(((d+1)<NbSymbols) && (boolean==FALSE)) {
          d++;
          if((Card[c][d]==TRUE) && (SymbolOccurence[d]<OccurencePerSymbol)) {
            //also check that this symbol fits with all other cards
            Card[NbCards][d]=TRUE;
            e=0; while((e<NbCards) && (CountIdenticalSymbols(e,NbCards)<=1)) { e++; };
            if(e==NbCards) { boolean=TRUE; }
            Card[NbCards][d]=FALSE;
          }
        }
        if(boolean==FALSE) { printf("error: no possible symbol\n"); PrintCards(); exit(-1); }
        
        Card[NbCards][d]=TRUE;
        SymbolOccurence[d]++;
      }
    }
    //count the nb of symbols set for this card and add more symbols if needed
    c=0;
    for(d=0;d<NbSymbols;d++) {
      if(Card[NbCards][d]==TRUE) { c++; }
    }
    if(c>SymbolsPerCard) { printf("error: too many symbols\n"); PrintCards(); exit(-1); }
    if(c<SymbolsPerCard) {
      for(d=0;d<SymbolsPerCard-c;d++) {
        Card[NbCards][NbSymbols+d]=TRUE;
        SymbolOccurence[NbSymbols+d]++;
      }
      NbSymbols+=SymbolsPerCard-c;
    }
    NbCards++;
//!!!    PrintCards();
  }
  
  printf("%d symbols per card, %d occurences per symbol\n",SymbolsPerCard,OccurencePerSymbol);
  printf("%d cards, %d symbols\n",NbCards,NbSymbols);
  PrintCards();
  
//!!!  printf("%d\n",AllSymbolsOccurCorrectly());
  
  exit(0);
}
